from dataclasses import dataclass, field
from ..core.model import DataModel

from typing import Optional

from .component_types import *

@dataclass
class SectionPart(DataModel, ContainerChild):
    """Represents the Section component."""

    accessory: SectionAccessoryChild = None
    """A component that is contextually associated to the content of the section."""

    components: list[SectionChild] = field(default_factory=list)
    """Component(s) representing the content of the section that is contextually associated to the accessory."""

    type: int = field(init=False, default=ComponentTypes.SECTION)
    """Component type. Always `ComponentTypes.SECTION` for this class. See [`ComponentTypes`][scurrypy.parts.component_types.ComponentTypes]."""

@dataclass
class TextDisplay(DataModel, ContainerChild, SectionChild):
    """Represents the Text Display component."""

    content: str = None
    """Text that will be displayed similar to a message."""

    type: int = field(init=False, default=ComponentTypes.TEXT_DISPLAY)
    """Component type. Always `ComponentTypes.TEXT_DISPLAY` for this class. See [`ComponentTypes`][scurrypy.parts.component_types.ComponentTypes]."""

@dataclass
class Thumbnail(DataModel, SectionAccessoryChild):
    """Represents the Thumbnail component."""
    
    media: str = None
    """Media of the thumbnail. http or attachment://<filename> scheme."""

    description: Optional[str] = None
    """Description for the media."""

    spoiler: Optional[bool] = False
    """Whether the thumbnail should be a spoiler (or blurred out). Defaults to `False`."""

    type: int = field(init=False, default=ComponentTypes.THUMBNAIL)
    """Component type. Always `ComponentTypes.THUMBNAIL` for this class. See [`ComponentTypes`][scurrypy.parts.component_types.ComponentTypes]."""

@dataclass
class MediaGalleryItem(DataModel):
    """Represents the Media Gallery Item component."""

    media: str = None
    """Image data. http or attachment://<filename> scheme."""

    description: Optional[str] = None
    """Alt text for the media."""

    spoiler: Optional[bool] = False
    """Whether the thumbnail should be a spoiler (or blurred out). Defaults to `False`."""

@dataclass
class MediaGallery(DataModel, ContainerChild):
    """Represents the Media Gallery component."""

    items: list[MediaGalleryItem] = field(default_factory=list)
    """1 to 10 nedia gallery items. See [`MediaGalleryItem`][scurrypy.parts.components_v2.MediaGalleryItem]."""

    type: int = field(init=False, default=ComponentTypes.MEDIA_GALLERY)
    """Component type. Always `ComponentTypes.MEDIA_GALLERY` for this class. See [`ComponentTypes`][scurrypy.parts.component_types.ComponentTypes]."""

@dataclass
class File(DataModel, ContainerChild):
    """Represents the File component."""

    file: str = None
    """File name. ONLY supports attachment://<filename> scheme."""

    spoiler: Optional[bool] = False
    """Whether the thumbnail should be a spoiler (or blurred out). Defaults to `False`."""

    type: int = field(init=False, default=ComponentTypes.FILE)
    """Component type. Always `ComponentTypes.FILE` for this class. See [`ComponentTypes`][scurrypy.parts.component_types.ComponentTypes]."""

class SeparatorTypes:
    """Represents separator types constants."""

    SMALL_PADDING = 1
    """Small separator padding."""
    
    LARGE_PADDING = 2
    """Large separator padding."""

@dataclass
class Separator(DataModel, ContainerChild):
    """Represents the Separator component."""

    divider: Optional[bool] = True
    """Whether a visual divider should be displayed in the component. Defaults to True."""

    spacing: Optional[int] = SeparatorTypes.SMALL_PADDING
    """Size of separator padding. Defaults to `SMALL_PADDING`. See [`SeparatorTypes`][scurrypy.parts.components_v2.SeparatorTypes]."""

    type: int = field(init=False, default=ComponentTypes.SEPARATOR)
    """Component type. Always `ComponentTypes.SEPARATOR` for this class. See [`ComponentTypes`][scurrypy.parts.component_types.ComponentTypes]."""

@dataclass
class ContainerPart(DataModel):
    """Represents a container of display and interactable components."""

    components: list[ContainerChild] = field(default_factory=list)
    """Child components that are encapsulated within the Container."""

    accent_color: Optional[int] = None
    """Color for the accent as an integer."""

    spoiler: Optional[bool] = False
    """If the container should be blurred out. Defaults to `False`."""

    type: int = field(init=False, default=ComponentTypes.CONTAINER)
    """Component type. Always `ComponentTypes.CONTAINER` for this class. See [`ComponentTypes`][scurrypy.parts.component_types.ComponentTypes]."""

@dataclass
class Label(DataModel):
    """Represents the Discord Label component."""

    label: str = None
    """Label text."""

    component: LabelChild = None
    """A component within the label."""

    description: Optional[str] = None
    """An optional description text for the label."""

    type: int = field(init=False, default=ComponentTypes.LABEL)
    """Component type. Always `ComponentTypes.LABEL` for this class. See [`ComponentTypes`][scurrypy.parts.component_types.ComponentTypes]."""

@dataclass
class FileUpload(DataModel, ContainerChild):
    """Represents the file upload component."""

    custom_id: str = None
    """ID for the file upload."""

    min_values: Optional[int] = 1
    """Minimum number of items that must be uploaded."""

    max_values: Optional[int] = 1
    """Maximum number of items that can be uploaded."""

    required: Optional[bool] = True
    """Whether files are required to be uploaded. Defaults to `True`."""

    type: int = field(init=False, default=ComponentTypes.FILE_UPLOAD)
    """Component type. Always `ComponentTypes.FILE_UPLOAD` for this class. See [`ComponentTypes`][scurrypy.parts.component_types.ComponentTypes]."""



class Addon:
    """Base class for writing addons.
    
    This class does not enforce structure; it only exists
    to give your addons a shared identity and future extension point.
    """
    ...



class EventTypes:
    """List of pre-defined event types."""
    
    # startup events
    READY = 'READY'

    # channel events
    CHANNEL_CREATE = 'CHANNEL_CREATE'
    CHANNEL_UPDATE = 'CHANNEL_UPDATE'
    CHANNEL_DELETE = 'CHANNEL_DELETE'
    CHANNEL_PINS_UPDATE = 'CHANNEL_PINS_UPDATE'

    # guild events
    GUILD_CREATE = 'GUILD_CREATE'
    GUILD_UPDATE = 'GUILD_UPDATE'
    GUILD_DELETE = 'GUILD_DELETE'
    GUILD_MEMBER_ADD = 'GUILD_MEMBER_ADD'
    GUILD_MEMBER_UPDATE = 'GUILD_MEMBER_UPDATE'
    GUILD_MEMBER_REMOVE = 'GUILD_MEMBER_REMOVE'
    GUILD_EMOJIS_UPDATE = 'GUILD_EMOJIS_UPDATE'

    # interaction events
    INTERACTION_CREATE = 'INTERACTION_CREATE'

    # message events
    MESSAGE_CREATE = 'MESSAGE_CREATE'
    MESSAGE_UPDATE = 'MESSAGE_UPDATE'
    MESSAGE_DELETE = 'MESSAGE_DELETE'

    # reaction events
    MESSAGE_REACTION_ADD = 'MESSAGE_REACTION_ADD'
    MESSAGE_REACTION_REMOVE = 'MESSAGE_REACTION_REMOVE'
    MESSAGE_REACTION_REMOVE_ALL = 'MESSAGE_REACTION_REMOVE_ALL'
    MESSAGE_REACTION_REMOVE_EMOJI = 'MESSAGE_REACTION_REMOVE_EMOJI'

    # role events
    ROLE_CREATE = 'ROLE_CREATE'
    ROLE_UPDATE = 'ROLE_UPDATE'
    ROLE_DELETE = 'ROLE_DELETE'



from dataclasses import dataclass
from typing import Unpack

from .base_resource import BaseResource

from ..parts.modal import ModalPart
from ..parts.message import MessagePart, MessageFlagParams, MessageFlags
from ..parts.command import CommandOptionChoice

from ..models.interaction import InteractionCallbackModel, InteractionCallbackTypes

@dataclass
class Interaction(BaseResource):
    """Represents a Discord Interaction object."""

    id: int
    """ID of the interaction."""

    token: str
    """Continuation token for responding to the interaction."""

    async def respond(self, message: str | MessagePart, with_response: bool = False, **flags: Unpack[MessageFlagParams]) -> InteractionCallbackModel | None:
        """Create a message in response to an interaction.

        Args:
            message (str | MessagePart): 2
            with_response (bool, optional): if the interaction data should be returned. Defaults to `False`.
            **flags: message flags to set. (set respective flag to `True` to toggle.)

        Raises:
            (TypeError): invalid `message` type

        Returns:
            (InteractionCallbackModel | None): interaction callback object (if `with_response` is toggled) else None
        """
        if isinstance(message, str):
            message = MessagePart(content=message).set_flags(**flags)
        elif not isinstance(message, MessagePart):
            raise TypeError(f"Interaction.respond expects type str or MessagePart, got {type(message).__name__}")
        
        content = {
            'type': InteractionCallbackTypes.CHANNEL_MESSAGE_WITH_SOURCE, 
            'data': message._prepare().to_dict()
        }
        
        data = await self._http.request(
            'POST', 
            f'/interactions/{self.id}/{self.token}/callback', 
            data=content, 
            files=[fp.path for fp in message.attachments],
            params={'with_response': with_response}
        )

        if with_response:
            return InteractionCallbackModel.from_dict(data)
        
    async def update(self, message: str | MessagePart) -> None:
        """Update a message in response to an interaction.

        Args:
            message (str | MessagePart): content as a string or MessagePart

        Raises:
            (TypeError): invalid `message` type
        """
        if isinstance(message, str):
            message = MessagePart(content=message)
        elif not isinstance(message, MessagePart):
            raise TypeError(f"Interaction.update expects type str or MessagePart, got {type(message).__name__}")
        
        content = {
            'type': InteractionCallbackTypes.UPDATE_MESSAGE, 
            'data': message._prepare().to_dict()
        }

        await self._http.request(
            'POST', 
            f'/interactions/{self.id}/{self.token}/callback', 
            data=content, 
            files=[fp.path for fp in message.attachments])

    async def respond_modal(self, modal: ModalPart) -> None:
        """Create a modal in response to an interaction.

        Args:
            modal (ModalPart): modal data

        Raises:
            (TypeError): invalid `modal` type
        """
        if not isinstance(modal, ModalPart):
            raise TypeError(f"Interaction.respond_modal expects type ModalPart, got {type(modal).__name__}")
        
        content = {
            'type': InteractionCallbackTypes.MODAL,
            'data': modal.to_dict()
        }

        await self._http.request(
            'POST', 
            f'/interactions/{self.id}/{self.token}/callback', 
            data=content)

    async def respond_autocomplete(self, choices: list[CommandOptionChoice]) -> None:
        """Autocomplete a command in response to an interaction.

        Args:
            choices (list[CommandOptionChoice]): list of choices to autocomplete
        """
        
        content = {
            'type': InteractionCallbackTypes.APPLICATION_COMMAND_AUTOCOMPLETE_RESULT,
            'data': {
                'choices': [choice.to_dict() for choice in choices]
            }
        }

        await self._http.request(
            'POST',
            f'/interactions/{self.id}/{self.token}/callback',
            data=content
        )

    async def defer_respond(self, ephemeral: bool) -> None:
        """Defer creating a message in response to an interaction.

        Args:
            ephemeral (bool): whether thinking + deferred interaction response is ephemeral
        """

        content = {
            'type': InteractionCallbackTypes.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE,
            'data': {
                'flags': MessageFlags.EPHEMERAL if ephemeral else 0
            }
        }

        await self._http.request(
            'POST',
            f'/interactions/{self.id}/{self.token}/callback',
            data=content
        )

    async def defer_update(self, ephemeral: bool) -> None:
        """Defer updating a message in response to an interaction.

        Args:
            ephemeral (bool): whether the deferred interaction response is ephemeral
        """

        content = {
            'type': InteractionCallbackTypes.DEFERRED_UPDATE_MESSAGE,
            'data': {
                'flags': MessageFlags.EPHEMERAL if ephemeral else 0
            }
        }

        await self._http.request(
            'POST',
            f'/interactions/{self.id}/{self.token}/callback',
            data=content
        )

    async def followup(self, application_id, message: str | MessagePart, **flags: Unpack[MessageFlagParams]) -> None:
        """Create a new message to respond to a deferred interaction.

        !!! important
            Apps are limited to 5 followup messages PER interaction.

        Args:
            application_id (int): ID of the application
            message (str | MessagePart): content as a string or MessagePart  
            **flags: message flags to set. (set respective flag to True to toggle.)

        Raises:
            (TypeError): invalid `message` type          
        """

        if isinstance(message, str):
            message = MessagePart(content=message).set_flags(**flags)
        elif not isinstance(message, MessagePart):
            raise TypeError(f"Interaction.respond expects type str or MessagePart, got {type(message).__name__}")
        
        content = message._prepare().to_dict()

        await self._http.request(
            'POST',
            f'/webhooks/{application_id}/{self.token}',
            data=content
        )

    async def edit_original(self, application_id: int, message: str | MessagePart) -> None:
        """Update the original interaction response from a deferred update interaction.

        Args:
            application_id (int): ID of the application
            message (str | MessagePart): content as a string or MessagePart

        Raises:
            (TypeError): invalid `message` type
        """

        if isinstance(message, str):
            message = MessagePart(content=message)
        elif not isinstance(message, MessagePart):
            raise TypeError(f"Interaction.respond expects type str or MessagePart, got {type(message).__name__}")
        
        content = message._prepare().to_dict()

        await self._http.request(
            'PATCH',
            f'/webhooks/{application_id}/{self.token}/messages/@original',
            data=content
        )
